/*
 * Copyright (c) 2005-2015 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <linux/joystick.h>

#include <neobotix/socketcan.h>
#include <neobotix/neo.h>
#include <neobotix/canGyro.h>
#include <neobotix/canIO.h>
#include <neobotix/canUS.h>
#include <neobotix/canMotors.h>

double rvel = 0.0;
double lvel = 0.0;
NEO_STR *neo;

/* For joystick driver */
#define NAME_LENGTH 128

unsigned char axes = 2;
unsigned char buttons = 2;
int version = 0x000800;
char name[NAME_LENGTH] = "Unknown";

/* Speed gains for joystick control */
#define VGAIN (-3.0/16384)	/* linear  */
#define WGAIN (-3.0/16384)	/* angular */

/*----------------------------------------------------------------------*/
/**
 ** Command task. Sends speed references to motors
 **/

static void*
watchdogTask(void *arg)
{
	struct timespec ts;

	if (canMotorsStart(neo) == -1) {
		warnx("Error canMotorStart");
		return NULL;
	}
	while (1) {
		if (canMotorsState(neo, CAN_MOTOR_LEFT)
		    == ST_OPERATION_ENABLED) 
			canMotorsSetVelGearRadS(neo, lvel, rvel);
		ts.tv_sec = 0;
		ts.tv_nsec = 100000000;
		nanosleep(&ts, NULL);
	}
	return NULL;
}

/*----------------------------------------------------------------------*/
/**
 ** Init joystick 
 **/
static int 
joyinit(char *dev)
{
	int joy;

	joy = open(dev, O_RDONLY);
	if (joy < 0) {
		perror(dev);
		return -1;
	}
	ioctl(joy, JSIOCGVERSION, &version);
	ioctl(joy, JSIOCGAXES, &axes);
	ioctl(joy, JSIOCGBUTTONS, &buttons);
	ioctl(joy, JSIOCGNAME(NAME_LENGTH), name);
	
	printf("Joystick (%s) has %d axes and %d buttons.\n"
	       "\tDriver version is %d.%d.%d.\n",
	       name, axes, buttons, version >> 16, 
	       (version >> 8) & 0xff, version & 0xff);
	return joy;
}
/*----------------------------------------------------------------------*/
/**
 ** Joystick control mode
 **/
static int
joymove(int fd)
{
	struct js_event js;
	int *axis;
	int *button;
	int ret = -1;

	axis = calloc(axes, sizeof(int));
	button = calloc(buttons, sizeof(int));

	printf("use button 4 to quit\n"
	       "   button 1 + left stick to move\n");
	while (1) {
		if (read(fd, &js, sizeof(struct js_event)) 
		    != sizeof(struct js_event)) {
			perror("\njstest: error reading");
			break;
		}
		
		switch(js.type & ~JS_EVENT_INIT) {
		case JS_EVENT_BUTTON:
			button[js.number] = js.value;
			printf("button %d %d\n", js.number, js.value);
			break;
		case JS_EVENT_AXIS:
			axis[js.number] = js.value;
			printf("axis %d %d\n", js.number, js.value);
			break;
		}
		/* Test for quit button */
		if (button[3]) {
			ret = 0;
			break;
		}
		/* Test for motion button */
		if (button[0]) {
			/* compute speed refs */
			rvel = axis[1]*VGAIN + axis[0]*WGAIN;
			lvel = -axis[1]*VGAIN + axis[0]*WGAIN;
			printf("move %lf %lf\n", lvel, rvel);
		} else {
			/* Set refs to zero */
			printf("stop\n");
			lvel = 0;
			rvel = 0;
		}
	} /* while */
	free(button);
	free(axis);
	return ret;
} /* joymove */

/*----------------------------------------------------------------------*/

/**
 ** Main 
 **/
int
main(int argc, char *argv[])
{
	double vel;
	char *devname;
	int m, joy = -1;
	pthread_t wdTaskId;

	if (argc > 2) 
		devname = argv[1];
	else
		devname = "can0";

	/* init neobotix control lib */
	neo = neoInit(devname);
	if (neo == NULL)
		err(2, "%s: neoInit failed", argv[0]);

	joy = joyinit("/dev/input/js0");
	if (joy < 0) {
		fprintf(stderr, "joyInit failed: %s\n", strerror(errno));
		neoEnd(neo);
		exit(2);
	}
	sleep(2);
	while (1) {
		printf(" 0 - quit            10 - stop US\n"		     
		       " 1 - connect gyro    11 - request US dist\n"	     
		       " 2 - request angle   12 - request US analog\n"	     
		       " 3 - show gyro	     13 - show US\n"		     
		       " 4 - zero gyro	     14 - start watchdog task\n"     
		       " 5 - connect IO	     15 - reset Motor Controllers\n" 
		       " 6 - request vbatt   16 - end watchdog task\n"	     
		       " 7 - show IO	     17 - set speed\n"		     
		       " 8 - connect US	     18 - show Motors\n"	     
		       " 9 - start US	     19 - joystick control\n"
		       "\n-> ");
		if (scanf("%d", &m) != 1)
			continue;
		switch (m) {
		case 0:
			printf("bye\n");
			neoEnd(neo);
			exit(0);
			break;
		case 1:
			canGyroRequestConnect(neo);
			break;
		case 2:
			canGyroRequestAngle(neo);
			break;
		case 3:
			canGyroShow(neo);
			break;
		case 4:
			canGyroZero(neo);
			break;
		case 5:
			canIORequestConnect(neo);
			break;
		case 6:
			// canIORequestStatus(neo);
			canIORequestVBatt(neo);
			canIORequestAnalogIn(neo);
			canIORequestDigiIn(neo);
			usleep(20000);
			canIOShow(neo);
			break;
		case 7:
			canIOShow(neo);
			break;
		case 8:
			canUSRequestConnect(neo);
			break;
		case 9:
			canUSStart(neo);
			break;
		case 10:
			canUSStop(neo);
			break;
		case 11:
			canUSRequestDist(neo);
			break;
		case 12:
			canUSRequestAnalog(neo);
			break;
		case 13:
			canUSShow(neo);
			break;
		case 14:
			pthread_create(&wdTaskId, NULL, watchdogTask, NULL);
			if (canMotorsReset(neo) == -1) 
				warnx("canMotorsReset");
			break;
			break;
		case 15:
			if (canMotorsReset(neo) == -1) 
				warnx("canMotorsReset");
			break;
		case 16:
			pthread_cancel(wdTaskId);
			break;
		case 17:
			printf("Enter speed: ");
			if (scanf("%lf", &vel) != 1)
				break;
			printf("vel: %f\n", vel);
			rvel = vel;
			lvel = -vel;
			break;
		case 18:
			canMotorsShow(neo);
			break;
		case 19:
			if (wdTaskId < 0) {
				printf("start watchdog task first!\n\n");
				break;
			}
			joymove(joy);
			break;

		default:
			printf("unkown command %d\n", m);
			break;
		}
	} /* while */
	neoEnd(neo);
}
